@@ -1,25 +1,28 @@
 package manifestcontroller
 
 import (
+	"context"
+	"encoding/json"
 	"fmt"
 	"testing"
 	"time"
 
-	fakeworkclient "github.com/open-cluster-management/api/client/work/clientset/versioned/fake"
-	workinformers "github.com/open-cluster-management/api/client/work/informers/externalversions"
-	workapiv1 "github.com/open-cluster-management/api/work/v1"
-	"github.com/open-cluster-management/work/pkg/spoke/controllers"
-	"github.com/open-cluster-management/work/pkg/spoke/spoketesting"
 	corev1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/equality"
 	"k8s.io/apimachinery/pkg/api/meta"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
 	"k8s.io/apimachinery/pkg/runtime"
+	"k8s.io/apimachinery/pkg/runtime/schema"
 	"k8s.io/apimachinery/pkg/util/diff"
 	fakedynamic "k8s.io/client-go/dynamic/fake"
 	fakekube "k8s.io/client-go/kubernetes/fake"
 	clienttesting "k8s.io/client-go/testing"
+	fakeworkclient "open-cluster-management.io/api/client/work/clientset/versioned/fake"
+	workinformers "open-cluster-management.io/api/client/work/informers/externalversions"
+	workapiv1 "open-cluster-management.io/api/work/v1"
+	"open-cluster-management.io/work/pkg/spoke/controllers"
+	"open-cluster-management.io/work/pkg/spoke/spoketesting"
 )
 
 type testController struct {
@@ -68,19 +71,9 @@ func (t *testController) withUnstructuredObject(objects ...runtime.Object) *test
 }
 
 func assertCondition(t *testing.T, conditions []metav1.Condition, expectedCondition string, expectedStatus metav1.ConditionStatus) {
-	conditionTypeFound := false
-	for _, condition := range conditions {
-		if condition.Type != expectedCondition {
-			continue
-		}
-		conditionTypeFound = true
-		if condition.Status != expectedStatus {
-			t.Errorf("expected %s but got: %s", expectedStatus, condition.Status)
-			break
-		}
-	}
+	matched := meta.IsStatusConditionPresentAndEqual(conditions, expectedCondition, expectedStatus)
 
-	if !conditionTypeFound {
+	if !matched {
 		t.Errorf("expected condition %s but got: %#v", expectedCondition, conditions)
 	}
 }
@@ -229,12 +222,13 @@ func (t *testCase) validate(
 	}
 }
 
-func newCondition(name, status, reason, message string, lastTransition *metav1.Time) metav1.Condition {
+func newCondition(name, status, reason, message string, generation int64, lastTransition *metav1.Time) metav1.Condition {
 	ret := metav1.Condition{
-		Type:    name,
-		Status:  metav1.ConditionStatus(status),
-		Reason:  reason,
-		Message: message,
+		Type:               name,
+		Status:             metav1.ConditionStatus(status),
+		ObservedGeneration: generation,
+		Reason:             reason,
+		Message:            message,
 	}
 	if lastTransition != nil {
 		ret.LastTransitionTime = *lastTransition
@@ -268,14 +262,14 @@ func TestSync(t *testing.T) {
 	cases := []*testCase{
 		newTestCase("create single resource").
 			withWorkManifest(spoketesting.NewUnstructured("v1", "Secret", "ns1", "test")).
-			withExpectedWorkAction("get", "update").
+			withExpectedWorkAction("update").
 			withAppliedWorkAction("create").
 			withExpectedKubeAction("get", "create").
 			withExpectedManifestCondition(expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionTrue}).
 			withExpectedWorkCondition(expectedCondition{string(workapiv1.WorkApplied), metav1.ConditionTrue}),
 		newTestCase("create single deployment resource").
 			withWorkManifest(spoketesting.NewUnstructured("apps/v1", "Deployment", "ns1", "test")).
-			withExpectedWorkAction("get", "update").
+			withExpectedWorkAction("update").
 			withAppliedWorkAction("create").
 			withExpectedDynamicAction("get", "create").
 			withExpectedManifestCondition(expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionTrue}).
@@ -283,14 +277,14 @@ func TestSync(t *testing.T) {
 		newTestCase("update single resource").
 			withWorkManifest(spoketesting.NewUnstructured("v1", "Secret", "ns1", "test")).
 			withSpokeObject(spoketesting.NewSecret("test", "ns1", "value2")).
-			withExpectedWorkAction("get", "update").
+			withExpectedWorkAction("update").
 			withAppliedWorkAction("create").
 			withExpectedKubeAction("get", "delete", "create").
 			withExpectedManifestCondition(expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionTrue}).
 			withExpectedWorkCondition(expectedCondition{string(workapiv1.WorkApplied), metav1.ConditionTrue}),
 		newTestCase("create single unstructured resource").
 			withWorkManifest(spoketesting.NewUnstructured("v1", "NewObject", "ns1", "test")).
-			withExpectedWorkAction("get", "update").
+			withExpectedWorkAction("update").
 			withAppliedWorkAction("create").
 			withExpectedDynamicAction("get", "create").
 			withExpectedManifestCondition(expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionTrue}).
@@ -298,7 +292,7 @@ func TestSync(t *testing.T) {
 		newTestCase("update single unstructured resource").
 			withWorkManifest(spoketesting.NewUnstructuredWithContent("v1", "NewObject", "ns1", "n1", map[string]interface{}{"spec": map[string]interface{}{"key1": "val1"}})).
 			withSpokeDynamicObject(spoketesting.NewUnstructuredWithContent("v1", "NewObject", "ns1", "n1", map[string]interface{}{"spec": map[string]interface{}{"key1": "val2"}})).
-			withExpectedWorkAction("get", "update").
+			withExpectedWorkAction("update").
 			withAppliedWorkAction("create").
 			withExpectedDynamicAction("get", "update").
 			withExpectedManifestCondition(expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionTrue}).
@@ -306,7 +300,7 @@ func TestSync(t *testing.T) {
 		newTestCase("multiple create&update resource").
 			withWorkManifest(spoketesting.NewUnstructured("v1", "Secret", "ns1", "test"), spoketesting.NewUnstructured("v1", "Secret", "ns2", "test")).
 			withSpokeObject(spoketesting.NewSecret("test", "ns1", "value2")).
-			withExpectedWorkAction("get", "update").
+			withExpectedWorkAction("update").
 			withAppliedWorkAction("create").
 			withExpectedKubeAction("get", "delete", "create", "get", "create").
 			withExpectedManifestCondition(expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionTrue}, expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionTrue}).
@@ -336,7 +330,7 @@ func TestFailedToApplyResource(t *testing.T) {
 	tc := newTestCase("multiple create&update resource").
 		withWorkManifest(spoketesting.NewUnstructured("v1", "Secret", "ns1", "test"), spoketesting.NewUnstructured("v1", "Secret", "ns2", "test")).
 		withSpokeObject(spoketesting.NewSecret("test", "ns1", "value2")).
-		withExpectedWorkAction("get", "update").
+		withExpectedWorkAction("update").
 		withAppliedWorkAction("create").
 		withExpectedKubeAction("get", "delete", "create", "get", "create").
 		withExpectedManifestCondition(expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionTrue}, expectedCondition{string(workapiv1.ManifestApplied), metav1.ConditionFalse}).
@@ -426,6 +420,7 @@ func TestGenerateUpdateStatusFunc(t *testing.T) {
 		name                     string
 		startingStatusConditions []metav1.Condition
 		manifestConditions       []workapiv1.ManifestCondition
+		generation               int64
 		expectedStatusConditions []metav1.Condition
 	}{
 		{
@@ -436,47 +431,49 @@ func TestGenerateUpdateStatusFunc(t *testing.T) {
 		{
 			name: "all manifests are applied successfully",
 			manifestConditions: []workapiv1.ManifestCondition{
-				newManifestCondition(0, "resource0", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(1, "resource1", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
+				newManifestCondition(0, "resource0", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(1, "resource1", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
 			},
 			expectedStatusConditions: []metav1.Condition{
-				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionTrue), "AppliedManifestWorkComplete", "Apply manifest work complete", nil),
+				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionTrue), "AppliedManifestWorkComplete", "Apply manifest work complete", 0, nil),
 			},
 		},
 		{
 			name: "one of manifests is not applied",
 			manifestConditions: []workapiv1.ManifestCondition{
-				newManifestCondition(0, "resource0", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(1, "resource1", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionFalse), "my-reason", "my-message", nil)),
+				newManifestCondition(0, "resource0", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(1, "resource1", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionFalse), "my-reason", "my-message", 0, nil)),
 			},
 			expectedStatusConditions: []metav1.Condition{
-				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionFalse), "AppliedManifestWorkFailed", "Failed to apply manifest work", nil),
+				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionFalse), "AppliedManifestWorkFailed", "Failed to apply manifest work", 0, nil),
 			},
 		},
 		{
 			name: "update existing status condition",
 			startingStatusConditions: []metav1.Condition{
-				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionTrue), "AppliedManifestWorkComplete", "Apply manifest work complete", &transitionTime),
+				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionTrue), "AppliedManifestWorkComplete", "Apply manifest work complete", 0, &transitionTime),
 			},
+			generation: 1,
 			manifestConditions: []workapiv1.ManifestCondition{
-				newManifestCondition(0, "resource0", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(1, "resource1", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
+				newManifestCondition(0, "resource0", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(1, "resource1", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
 			},
 			expectedStatusConditions: []metav1.Condition{
-				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionTrue), "AppliedManifestWorkComplete", "Apply manifest work complete", &transitionTime),
+				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionTrue), "AppliedManifestWorkComplete", "Apply manifest work complete", 1, &transitionTime),
 			},
 		},
 		{
 			name: "override existing status conditions",
 			startingStatusConditions: []metav1.Condition{
-				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionTrue), "AppliedManifestWorkComplete", "Apply manifest work complete", nil),
+				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionTrue), "AppliedManifestWorkComplete", "Apply manifest work complete", 0, nil),
 			},
 			manifestConditions: []workapiv1.ManifestCondition{
-				newManifestCondition(0, "resource0", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(1, "resource1", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionFalse), "my-reason", "my-message", nil)),
+				newManifestCondition(0, "resource0", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(1, "resource1", newCondition(string(workapiv1.ManifestApplied), string(metav1.ConditionFalse), "my-reason", "my-message", 0, nil)),
 			},
+			generation: 1,
 			expectedStatusConditions: []metav1.Condition{
-				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionFalse), "AppliedManifestWorkFailed", "Failed to apply manifest work", nil),
+				newCondition(string(workapiv1.WorkApplied), string(metav1.ConditionFalse), "AppliedManifestWorkFailed", "Failed to apply manifest work", 1, nil),
 			},
 		},
 	}
@@ -484,7 +481,7 @@ func TestGenerateUpdateStatusFunc(t *testing.T) {
 	controller := &ManifestWorkController{}
 	for _, c := range cases {
 		t.Run(c.name, func(t *testing.T) {
-			updateStatusFunc := controller.generateUpdateStatusFunc(c.manifestConditions)
+			updateStatusFunc := controller.generateUpdateStatusFunc(c.generation, c.manifestConditions)
 			manifestWorkStatus := &workapiv1.ManifestWorkStatus{
 				Conditions: c.startingStatusConditions,
 			}
@@ -518,8 +515,8 @@ func TestAllInCondition(t *testing.T) {
 			name:          "condition does not exist",
 			conditionType: "one",
 			manifestConditions: []workapiv1.ManifestCondition{
-				newManifestCondition(0, "resource0", newCondition("two", string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(1, "resource1", newCondition("two", string(metav1.ConditionFalse), "my-reason", "my-message", nil)),
+				newManifestCondition(0, "resource0", newCondition("two", string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(1, "resource1", newCondition("two", string(metav1.ConditionFalse), "my-reason", "my-message", 0, nil)),
 			},
 			expected: []bool{false, false},
 		},
@@ -527,10 +524,10 @@ func TestAllInCondition(t *testing.T) {
 			name:          "all manifests are in the condition",
 			conditionType: "one",
 			manifestConditions: []workapiv1.ManifestCondition{
-				newManifestCondition(0, "resource0", newCondition("one", string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(1, "resource1", newCondition("one", string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(2, "resource0", newCondition("two", string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(3, "resource1", newCondition("two", string(metav1.ConditionFalse), "my-reason", "my-message", nil)),
+				newManifestCondition(0, "resource0", newCondition("one", string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(1, "resource1", newCondition("one", string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(2, "resource0", newCondition("two", string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(3, "resource1", newCondition("two", string(metav1.ConditionFalse), "my-reason", "my-message", 0, nil)),
 			},
 			expected: []bool{true, true},
 		},
@@ -538,10 +535,10 @@ func TestAllInCondition(t *testing.T) {
 			name:          "one of manifests is not in the condition",
 			conditionType: "two",
 			manifestConditions: []workapiv1.ManifestCondition{
-				newManifestCondition(0, "resource0", newCondition("one", string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(1, "resource1", newCondition("one", string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(2, "resource0", newCondition("two", string(metav1.ConditionTrue), "my-reason", "my-message", nil)),
-				newManifestCondition(3, "resource1", newCondition("two", string(metav1.ConditionFalse), "my-reason", "my-message", nil)),
+				newManifestCondition(0, "resource0", newCondition("one", string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(1, "resource1", newCondition("one", string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(2, "resource0", newCondition("two", string(metav1.ConditionTrue), "my-reason", "my-message", 0, nil)),
+				newManifestCondition(3, "resource1", newCondition("two", string(metav1.ConditionFalse), "my-reason", "my-message", 0, nil)),
 			},
 			expected: []bool{false, true},
 		},
@@ -612,7 +609,7 @@ func TestBuildResourceMeta(t *testing.T) {
 
 	for _, c := range cases {
 		t.Run(c.name, func(t *testing.T) {
-			actual, err := buildResourceMeta(0, c.object, c.restMapper)
+			actual, _, err := buildResourceMeta(0, c.object, c.restMapper)
 			if err != nil {
 				t.Errorf("Should be success with no err: %v", err)
 			}
@@ -633,12 +630,6 @@ func TestBuildManifestResourceMeta(t *testing.T) {
 		restMapper     meta.RESTMapper
 		expected       workapiv1.ManifestResourceMeta
 	}{
-		{
-			name:        "extract meta from apply result",
-			applyResult: spoketesting.NewSecret("test1", "ns1", "value1"),
-			restMapper:  spoketesting.NewFakeRestMapper(),
-			expected:    workapiv1.ManifestResourceMeta{Version: "v1", Kind: "Secret", Resource: "secrets", Namespace: "ns1", Name: "test1"},
-		},
 		{
 			name:           "fall back to manifest",
 			manifestObject: spoketesting.NewSecret("test2", "ns2", "value2"),
@@ -653,7 +644,7 @@ func TestBuildManifestResourceMeta(t *testing.T) {
 			if c.manifestObject != nil {
 				manifest.Object = c.manifestObject
 			}
-			actual, err := buildManifestResourceMeta(0, c.applyResult, manifest, c.restMapper)
+			actual, _, err := buildManifestResourceMeta(0, manifest, c.restMapper)
 			if err != nil {
 				t.Errorf("Should be success with no err: %v", err)
 			}
@@ -665,3 +656,223 @@ func TestBuildManifestResourceMeta(t *testing.T) {
 		})
 	}
 }
+
+func TestManageOwner(t *testing.T) {
+	testGVR := schema.GroupVersionResource{Version: "v1", Resource: "secrets"}
+
+	namespace := "testns"
+
+	name := "test"
+
+	cases := []struct {
+		name         string
+		deleteOption *workapiv1.DeleteOption
+		owner        metav1.OwnerReference
+		expectOwner  metav1.OwnerReference
+	}{
+		{
+			name:        "foreground by default",
+			owner:       metav1.OwnerReference{UID: "testowner"},
+			expectOwner: metav1.OwnerReference{UID: "testowner"},
+		},
+		{
+			name:         "orphan the resource",
+			owner:        metav1.OwnerReference{UID: "testowner"},
+			deleteOption: &workapiv1.DeleteOption{PropagationPolicy: workapiv1.DeletePropagationPolicyTypeOrphan},
+			expectOwner:  metav1.OwnerReference{UID: "testowner-"},
+		},
+		{
+			name:         "add owner if no orphan rule with selectively orphan",
+			owner:        metav1.OwnerReference{UID: "testowner"},
+			deleteOption: &workapiv1.DeleteOption{PropagationPolicy: workapiv1.DeletePropagationPolicyTypeSelectivelyOrphan},
+			expectOwner:  metav1.OwnerReference{UID: "testowner"},
+		},
+		{
+			name:  "orphan the resource with selectively orphan",
+			owner: metav1.OwnerReference{UID: "testowner"},
+			deleteOption: &workapiv1.DeleteOption{
+				PropagationPolicy: workapiv1.DeletePropagationPolicyTypeSelectivelyOrphan,
+				SelectivelyOrphan: &workapiv1.SelectivelyOrphan{
+					OrphaningRules: []workapiv1.OrphaningRule{
+						{
+							Group:     "",
+							Resource:  "secrets",
+							Namespace: namespace,
+							Name:      name,
+						},
+					},
+				},
+			},
+			expectOwner: metav1.OwnerReference{UID: "testowner-"},
+		},
+		{
+			name:  "add owner if resourcec is not matched in orphan rule with selectively orphan",
+			owner: metav1.OwnerReference{UID: "testowner"},
+			deleteOption: &workapiv1.DeleteOption{
+				PropagationPolicy: workapiv1.DeletePropagationPolicyTypeSelectivelyOrphan,
+				SelectivelyOrphan: &workapiv1.SelectivelyOrphan{
+					OrphaningRules: []workapiv1.OrphaningRule{
+						{
+							Group:     "",
+							Resource:  "secrets",
+							Namespace: "testns1",
+							Name:      name,
+						},
+					},
+				},
+			},
+			expectOwner: metav1.OwnerReference{UID: "testowner"},
+		},
+	}
+
+	for _, c := range cases {
+		t.Run(c.name, func(t *testing.T) {
+			owner := manageOwnerRef(testGVR, namespace, name, c.deleteOption, c.owner)
+
+			if !equality.Semantic.DeepEqual(owner, c.expectOwner) {
+				t.Errorf("Expect owner is %v, but got %v", c.expectOwner, owner)
+			}
+		})
+	}
+}
+
+func TestApplyUnstructred(t *testing.T) {
+	cases := []struct {
+		name            string
+		owner           metav1.OwnerReference
+		existingObject  []runtime.Object
+		required        *unstructured.Unstructured
+		gvr             schema.GroupVersionResource
+		validateActions func(t *testing.T, actions []clienttesting.Action)
+	}{
+		{
+			name:           "create a new object with owner",
+			existingObject: []runtime.Object{},
+			owner:          metav1.OwnerReference{Name: "test", UID: "testowner"},
+			required:       spoketesting.NewUnstructured("v1", "Secret", "ns1", "test"),
+			gvr:            schema.GroupVersionResource{Version: "v1", Resource: "secrets"},
+			validateActions: func(t *testing.T, actions []clienttesting.Action) {
+				if len(actions) != 2 {
+					t.Errorf("Expect 2 actions, but have %d", len(actions))
+				}
+
+				spoketesting.AssertAction(t, actions[0], "get")
+				spoketesting.AssertAction(t, actions[1], "create")
+
+				obj := actions[1].(clienttesting.CreateActionImpl).Object.(*unstructured.Unstructured)
+				owners := obj.GetOwnerReferences()
+				if len(owners) != 1 {
+					t.Errorf("Expect 2 owners, but have %d", len(owners))
+				}
+
+				if owners[0].UID != "testowner" {
+					t.Errorf("Owner UId is not correct, got %s", owners[0].UID)
+				}
+			},
+		},
+		{
+			name:           "create a new object without owner",
+			existingObject: []runtime.Object{},
+			owner:          metav1.OwnerReference{Name: "test", UID: "testowner-"},
+			required:       spoketesting.NewUnstructured("v1", "Secret", "ns1", "test"),
+			gvr:            schema.GroupVersionResource{Version: "v1", Resource: "secrets"},
+			validateActions: func(t *testing.T, actions []clienttesting.Action) {
+				if len(actions) != 2 {
+					t.Errorf("Expect 2 actions, but have %d", len(actions))
+				}
+
+				spoketesting.AssertAction(t, actions[0], "get")
+				spoketesting.AssertAction(t, actions[1], "create")
+
+				obj := actions[1].(clienttesting.CreateActionImpl).Object.(*unstructured.Unstructured)
+				owners := obj.GetOwnerReferences()
+				if len(owners) != 0 {
+					t.Errorf("Expect 1 owners, but have %d", len(owners))
+				}
+			},
+		},
+		{
+			name:           "update an object owner",
+			existingObject: []runtime.Object{spoketesting.NewUnstructured("v1", "Secret", "ns1", "test", metav1.OwnerReference{Name: "test1", UID: "testowner1"})},
+			owner:          metav1.OwnerReference{Name: "test", UID: "testowner"},
+			required:       spoketesting.NewUnstructured("v1", "Secret", "ns1", "test"),
+			gvr:            schema.GroupVersionResource{Version: "v1", Resource: "secrets"},
+			validateActions: func(t *testing.T, actions []clienttesting.Action) {
+				if len(actions) != 2 {
+					t.Errorf("Expect 2 actions, but have %d", len(actions))
+				}
+
+				spoketesting.AssertAction(t, actions[0], "get")
+				spoketesting.AssertAction(t, actions[1], "update")
+
+				obj := actions[1].(clienttesting.UpdateActionImpl).Object.(*unstructured.Unstructured)
+				owners := obj.GetOwnerReferences()
+				if len(owners) != 2 {
+					t.Errorf("Expect 2 owners, but have %d", len(owners))
+				}
+
+				if owners[0].UID != "testowner1" {
+					t.Errorf("Owner UId is not correct, got %s", owners[0].UID)
+				}
+				if owners[1].UID != "testowner" {
+					t.Errorf("Owner UId is not correct, got %s", owners[1].UID)
+				}
+			},
+		},
+		{
+			name:           "update an object without owner",
+			existingObject: []runtime.Object{spoketesting.NewUnstructured("v1", "Secret", "ns1", "test", metav1.OwnerReference{Name: "test1", UID: "testowner1"})},
+			owner:          metav1.OwnerReference{Name: "test", UID: "testowner-"},
+			required:       spoketesting.NewUnstructured("v1", "Secret", "ns1", "test"),
+			gvr:            schema.GroupVersionResource{Version: "v1", Resource: "secrets"},
+			validateActions: func(t *testing.T, actions []clienttesting.Action) {
+				if len(actions) != 1 {
+					t.Errorf("Expect 1 actions, but have %d", len(actions))
+				}
+
+				spoketesting.AssertAction(t, actions[0], "get")
+			},
+		},
+		{
+			name:           "remove an object owner",
+			existingObject: []runtime.Object{spoketesting.NewUnstructured("v1", "Secret", "ns1", "test", metav1.OwnerReference{Name: "test", UID: "testowner"})},
+			owner:          metav1.OwnerReference{Name: "test", UID: "testowner-"},
+			required:       spoketesting.NewUnstructured("v1", "Secret", "ns1", "test"),
+			gvr:            schema.GroupVersionResource{Version: "v1", Resource: "secrets"},
+			validateActions: func(t *testing.T, actions []clienttesting.Action) {
+				if len(actions) != 2 {
+					t.Errorf("Expect 2 actions, but have %d", len(actions))
+				}
+
+				spoketesting.AssertAction(t, actions[0], "get")
+				spoketesting.AssertAction(t, actions[1], "update")
+
+				obj := actions[1].(clienttesting.UpdateActionImpl).Object.(*unstructured.Unstructured)
+				owners := obj.GetOwnerReferences()
+				if len(owners) != 0 {
+					t.Errorf("Expect 0 owner, but have %d", len(owners))
+				}
+			},
+		},
+	}
+
+	for _, c := range cases {
+		t.Run(c.name, func(t *testing.T) {
+			work, workKey := spoketesting.NewManifestWork(0)
+			work.Finalizers = []string{controllers.ManifestWorkFinalizer}
+			controller := newController(work, nil, spoketesting.NewFakeRestMapper()).
+				withUnstructuredObject(c.existingObject...)
+			syncContext := spoketesting.NewFakeSyncContext(t, workKey)
+
+			data, _ := json.Marshal(c.required)
+			_, _, err := controller.controller.applyUnstructured(
+				context.TODO(), data, c.owner, c.gvr, syncContext.Recorder())
+
+			if err != nil {
+				t.Errorf("expect no error, but got %v", err)
+			}
+
+			c.validateActions(t, controller.dynamicClient.Actions())
+		})
+	}
+}
